package com.segment.analytics;

import android.util.Base64;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.URL;

/**
 * Abstraction to customize how connections are created. This is can be used to point our SDK at
 * your proxy server for instance.
 */
public class ConnectionFactory {

  private static final int DEFAULT_READ_TIMEOUT_MILLIS = 20 * 1000; // 20s
  private static final int DEFAULT_CONNECT_TIMEOUT_MILLIS = 15 * 1000; // 15s

  // NOTE: "writeKey" MUST BE in the format "Stage-v2-aBc123 ..."
  //    where "Stage" is the API stage ("Dev", "Prod", etc.),
  //          "v2"    is the API version,
  //          "aBc123 ..." is a random string.
  // Only alphanumeric ASCII characters are allowed (besides
  // the two separator dashes).
  private String[] splitWriteKey(String writeKey) {
    return writeKey.split("-");
  }

  /** Return a {@link HttpURLConnection} that reads JSON formatted project settings. */
  public HttpURLConnection projectSettings(String writeKey) throws IOException {
    String[] keyComponents = splitWriteKey(writeKey);

    // WARNING: the "writeKey" is not URL encoded, or checked for disallowed
    // characters.  This was the case with Segment's original code also.
    HttpURLConnection connection = openConnection(
            String.format("https://analytics.kidoodle.tv/%s/%s/projects/%s/settings",
                    keyComponents[1], keyComponents[0], writeKey
            )
      );
    connection.setRequestProperty("X-Api-Key", writeKey);
    return connection;
  }

  /**
   * Return a {@link HttpURLConnection} that writes batched payloads to {@code
   * https://analytics.kidoodle.tv/VERSION/STAGE/import}.
   */
  public HttpURLConnection upload(String writeKey) throws IOException {
    String[] keyComponents = splitWriteKey(writeKey);

    // WARNING: the "writeKey" is not URL encoded, or checked for disallowed
    // characters.  This was the case with Segment's original code also.
    HttpURLConnection connection = openConnection(
	    String.format("https://analytics.kidoodle.tv/%s/%s/import",
		  keyComponents[1], keyComponents[0]
	    )
    );
    connection.setRequestProperty("X-Api-Key", writeKey);
    connection.setRequestProperty("Content-Encoding", "gzip");
    connection.setDoOutput(true);
    connection.setChunkedStreamingMode(0);
    return connection;
  }


  /**
   * Configures defaults for connections opened with {@link #upload(String)}
   * and {@link #projectSettings(String)}.
   */
  protected HttpURLConnection openConnection(String url) throws IOException {
    HttpURLConnection connection = (HttpURLConnection) new URL(url).openConnection();
    connection.setConnectTimeout(DEFAULT_CONNECT_TIMEOUT_MILLIS);
    connection.setReadTimeout(DEFAULT_READ_TIMEOUT_MILLIS);
    connection.setRequestProperty("Content-Type", "application/json");
    connection.setDoInput(true);
    return connection;
  }
}
